"""
Processes the AqSolDB dataset for the water solubility of chemical compounds from the raw format as it is
provided by the original authors into a format that can be used by other experiments. The datasets consists
of ~10000 molecules annotated with measurements of water solubility at room temperatures which are to be
predicted. The original format of the dataset is a CSV file of SMILES strings of the molecules. Further
experiments however need it in the format of a so called "eye tracking dataset". In this format the datasets
consists of a folder, where every element is associated with 2 files: (1) An image which contains a
visualization of the corresponding molecule and (2) a json metadata file. The metadata file has to contain
the target solubility value and an appropriate graph structure representing the molecule, which contains
numeric feature vectors for each node and each graph. This processing is done by using RDKit to interpret
and draw the molecules based on the SMILES string and then export the graph representation using some of
atom and bond properties calculated by RDkit.
"""
import os
import sys
import csv
import math
import pathlib
import random
from zipfile import ZipFile
from itertools import chain
from typing import Optional, List

import numpy as np
import matplotlib.pyplot as plt
from kgcnn.data.moleculenet import MoleculeNetDataset, OneHotEncoder

from pycomex.experiment import Experiment
from pycomex.util import Skippable

# from gnn_teacher_student.util import graph_dict_to_list_values
# from gnn_teacher_student.visualization import pdf_from_eye_tracking_dataset
from graph_attention_student.data import create_molecule_eye_tracking_dataset, load_eye_tracking_dataset
from graph_attention_student.util import DATASETS_FOLDER

SHORT_DESCRIPTION = (
    'Processes the raw AqSolDB dataset for water solubility of molucules into a format which can be used by '
    'the other experiments.'
)

PATH = os.path.dirname(pathlib.Path(__file__).parent.absolute())
BASE_PATH = os.getenv('EXPERIMENT_BASE_PATH', os.path.join(PATH, 'experiment_results'))

# == SOURCE PARAMETERS ==
TRAIN_DATASET_PATH = os.path.join(DATASETS_FOLDER, 'aqsoldb_raw', 'dataset-all.csv')
TEST_DATASET_PATH = os.path.join(DATASETS_FOLDER, 'aqsoldb_raw', 'dataset-E.csv')

# == PROCESSING PARAMETERS ==
ATOMS = ['Cl', 'C', 'N', 'O', 'F', 'S', 'Br', 'I', 'H', 'Na', 'P', 'Ca', 'B']
# The dimensions which the images of the molecular graphs are supposed to have in the end, given in pixels
WIDTH, HEIGHT = 900, 900

# == EVALUATION PARAMETERS ==

def plot_distribution(ax: plt.Axes,
                      values: List[float],
                      bins: List[int],
                      color='coral'):

    hist, bin_edges = np.histogram(values, bins)
    xs = list(range(len(hist)))
    ax.bar(xs, hist, color=color)
    ax.set_xticks(xs)
    ax.set_xticklabels([f'[{bin_edges[i]:.1f}, {bin_edges[i + 1]:.1f})' for i in range(len(bin_edges) - 1)])
    ax.set_ylabel('Count')


def molecule_filter(mg, ds):
    data = dict(ds)
    # The paper which acts as the main comparison lists some pre-processing steps which we will recreate
    # here. A compound will not be used if...

    # There is no carbon atom in it at all
    if 'C' not in data['SMILES']:
        return None

    # The compound contains charged atoms
    if '+' in data['SMILES'] or '-' in data['SMILES']:
        return None

    # If the element consists of adjoined mixtures
    if '.' in data['SMILES']:
        return None

    return 1


def load_moleculenet(csv_path: str,
                     name: str = 'aqsoldb',
                     name_blacklist: List[str] = []):
    moleculenet = MoleculeNetDataset(
        file_name=os.path.basename(csv_path),
        data_directory=os.path.dirname(csv_path),
        dataset_name=name
    )

    moleculenet.prepare_data(
        overwrite=False,
        smiles_column_name='SMILES',
        add_hydrogen=True,
        make_conformers=True,
        optimize_conformer=True
    )

    moleculenet.read_in_memory(
        label_column_name=['Solubility'],
        add_hydrogen=False,
        has_conformers=True
    )

    moleculenet.set_attributes(
        nodes=['Symbol', 'TotalNumHs'],
        encoder_nodes={
            'Symbol': OneHotEncoder(ATOMS, dtype='str', add_unknown=False),
            'TotalNumHs': int
        },
        edges=['BondType'],
        encoder_edges={
            'BondType': int
        },
        graph=['NumAtoms'],
        additional_callbacks={
            'node_count': lambda mg, ds: int(mg.mol.GetNumAtoms()),
            'solubility': lambda mg, ds: float(dict(ds)['Solubility']),
            'smiles': lambda mg, ds: str(dict(ds)['SMILES']),
            'id': lambda mg, ds: str(dict(ds)['ID']),
            'name': lambda mg, ds: str(dict(ds)['Name']),
            # AqSolDB consists of multiple datasets, which are enumerated with characters A, B, C, D...
            # which are encoded into the ID of each element and here we retrieve it, because that will
            # later be important to determine the test set.
            'origin': lambda mg, ds: str(dict(ds)['ID']).split('-')[0].upper(),
            # This is a boolean value which is None if we do not want to include that compound into the
            # dataset.
            'filter': molecule_filter,
            'name_filter': lambda mg, ds: None if str(dict(ds)['Name']) in name_blacklist else 1
        }
    )
    moleculenet.clean(['filter', 'name_filter', 'edge_attributes'])
    return moleculenet


NAMESPACE = 'process_aqsoldb'
BASE_PATH = os.getcwd()
DEBUG = True

with Skippable(), (e := Experiment(base_path=BASE_PATH, namespace=NAMESPACE, glob=globals())):

    # -- Loading the original dataset from CSV
    e.info('Loading CSV file...')
    # For each line in the CSV a dict will be added to this list
    raw_dataset: List[dict] = []
    with open(TRAIN_DATASET_PATH, mode='r') as csv_file:
        reader = csv.DictReader(csv_file, delimiter=',', quotechar='"')
        for row in reader:
            raw_dataset.append(row)

    e.info(f'{len(raw_dataset)} rows of data')
    e.info(f'{len(reader._fieldnames)} fields: {", ".join(reader._fieldnames)}')
    e.info(f'printing one example:')
    for key, value in raw_dataset[0].items():
        e.info(f' * {key:<20}: {value}')

    # -- Processing the dataset with KGCNN
    test_moleculenet = load_moleculenet(TEST_DATASET_PATH)
    e.info(f'loaded test dataset with {len(test_moleculenet)} elements')

    test_names = [g['name'] for g in test_moleculenet]
    train_moleculenet = load_moleculenet(TRAIN_DATASET_PATH, name_blacklist=test_names)
    e.info(f'loaded train dataset with {len(train_moleculenet)} elements')

    # -- Actually creating the eye tracking dataset
    dataset_path = os.path.join(e.path, 'dataset')
    os.mkdir(dataset_path)

    infos = {}
    for dataset_type, dataset in [('train', train_moleculenet), ('test', test_moleculenet)]:
        for g in dataset:
            infos[str(g['id'])] = {
                'smiles': str(g['smiles']),
                'id': str(g['id']),
                'input_type': 'regression',
                'solubility': float(g['solubility']),
                'value': float(g['solubility']),
                'name': str(g['name']),
                'type': dataset_type
            }

    create_molecule_eye_tracking_dataset(
        molecule_infos=infos,
        dest_path=dataset_path,
        image_width=WIDTH,
        image_height=HEIGHT,
        logger=e.logger,
        set_attributes_kwargs={
            'nodes': ['Symbol', 'TotalDegree', 'FormalCharge', 'NumRadicalElectrons', 'Hybridization',
                      'IsAromatic', 'IsInRing', 'TotalNumHs', 'CIPCode', 'ChiralityPossible', 'ChiralTag'],
            'encoder_nodes': {
                'Symbol': OneHotEncoder(
                    ['B', 'C', 'N', 'O', 'F', 'Si', 'P', 'S', 'Cl', 'As', 'Se', 'Br', 'Te', 'I', 'At'],
                    dtype="str"
                ),
                'Hybridization': OneHotEncoder([2, 3, 4, 5, 6]),
                'TotalDegree': OneHotEncoder([0, 1, 2, 3, 4, 5], add_unknown=False),
                'TotalNumHs': OneHotEncoder([0, 1, 2, 3, 4], add_unknown=False),
                'CIPCode': OneHotEncoder(['R', 'S'], add_unknown=False, dtype='str'),
                'ChiralityPossible': OneHotEncoder(["1"], add_unknown=False, dtype='str'),
            },
            'edges': ['BondType', 'IsAromatic', 'IsConjugated', 'IsInRing', 'Stereo'],
            'encoder_edges': {
                'BondType': OneHotEncoder([1, 2, 3, 12], add_unknown=False),
                'Stereo': OneHotEncoder([0, 1, 2, 3], add_unknown=False)
            }
        }
    )

with Skippable(), e.analysis:
    # The previous create command only creates the dataset files, here we load it into memory as a dict
    # structure so that we can create the overview pdf from this.
    # "pdf_from_eye_tracking_dataset" creates a PDF file which contains some general information about the
    # dataset as well as all the images and the importance ground truth illustrations!
    eye_tracking_dataset = load_eye_tracking_dataset(dataset_path)
    pass







"""
This example shows how to train a KGCNN GIN model for multitask graph regression application based on an
existing "visual graph dataset".

Inherits "vgd_multitask_gnn.py"

**CHANGELOG**

0.1.0 - 18.01.2022 - Initial version
"""
import os
import warnings
import pathlib
warnings.filterwarnings("ignore")
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import random
import typing as t

import numpy as np
import matplotlib.pyplot as plt
import tensorflow as tf
import tensorflow.keras as ks
from pycomex.experiment import SubExperiment
from pycomex.util import Skippable
from kgcnn.layers.conv.gat_conv import AttentionHeadGATV2
from kgcnn.layers.mlp import GraphMLP
from kgcnn.layers.conv.gin_conv import GIN
from kgcnn.layers.modules import LazyConcatenate, DenseEmbedding, DropoutEmbedding
from kgcnn.layers.pooling import PoolingGlobalEdges

# local
from graph_attention_student.training import mse


# == DATASET PARAMETERS ==

USE_GRAPH_ATTRIBUTES: bool = True
USE_EDGE_FEATURES: bool = False


# == MODEL PARAMETERS ==

class GinModel(ks.models.Model):

    def __init__(self,
                 units: t.List[t.List[int]],
                 final_units: t.List[int],
                 pooling_method: str = 'sum',
                 activation: str = 'kgcnn>leaky_relu',
                 final_pooling: str = 'sum',
                 final_activation: str = 'linear',
                 dropout_rate: float = 0.0,
                 use_graph_attributes: bool = False):
        super(GinModel, self).__init__()
        self.use_graph_attributes = use_graph_attributes

        self.lay_dropout = DropoutEmbedding(rate=dropout_rate)
        self.conv_layers: t.List[t.Tuple[ks.layers.Layer, ks.layers.Layer]] = []
        for k in units:
            lay = GIN(
                pooling_method=pooling_method,
                epsilon_learnable=True,
            )
            mlp = GraphMLP(
                units=k,
                activation=activation,
                use_dropout=True,
                rate=dropout_rate,
            )
            self.conv_layers.append((lay, mlp))

        self.lay_pooling = PoolingGlobalEdges(pooling_method=final_pooling)
        self.lay_concat = LazyConcatenate(axis=-1)

        self.final_layers = []
        self.final_activations = ['relu' for _ in final_units]
        self.final_activations[-1] = final_activation
        for k, act in zip(final_units, self.final_activations):
            lay = DenseEmbedding(units=k, activation=act)
            self.final_layers.append(lay)

    def call(self, inputs, training=False):
        if self.use_graph_attributes:
            node_input, edge_input, edge_index_input, graph_input = inputs
        else:
            node_input, edge_input, edge_index_input = inputs

        x = node_input
        for lay, mlp in self.conv_layers:
            x = lay([x, edge_index_input])
            x = mlp(x)
            if training:
                x = self.lay_dropout(x)

        final = self.lay_pooling(x)
        if self.use_graph_attributes:
            final = self.lay_concat([final, graph_input])

        for lay in self.final_layers:
            final = lay(final)

        return final


MODEL_NAME = 'GIN'
DROPOUT_RATE = 0.1
UNITS = [[32, 32], [32, 32], [32, 32]]

# == TRAINING PARAMETERS ==
LEARNING_RATE = 0.001
EPOCHS: int = 1000

# == EXPERIMENT PARAMETERS ==
PATH = pathlib.Path(__file__).parent.absolute()
EXPERIMENT_PATH = os.path.join(PATH, 'vgd_multitask_gnn.py')
BASE_PATH = os.getcwd()
NAMESPACE = 'results/vgd_multitask_gnn_gin'
DEBUG = True
with Skippable(), (se := SubExperiment(EXPERIMENT_PATH, BASE_PATH, NAMESPACE, globals())):

    # Here we need to overwrite this hook in case we want ot use the edge features. In the original
    # experiment, the implementation of this hook removes the edge features from the dataset, which
    # means by overwriting it we would prevent that and allow the edge features to be part of the dataset
    @se.hook('modify_g', replace=USE_EDGE_FEATURES)
    def allow_edge_features(e, g):
        return g

    @se.hook('create_model')
    def create_model(e):
        e.info('creating the GATv2 model...')
        e.info(f' * UNITS: {e.parameters["UNITS"]}')
        e.info(f' * FINAL_UNITS: {e.parameters["FINAL_UNITS"]}')
        e.info(f' * DROPOUT_RATE: {e.parameters["DROPOUT_RATE"]}')

        model = GinModel(
            units=e.parameters['UNITS'],
            final_units=e.parameters['FINAL_UNITS'],
            use_graph_attributes=e.parameters['USE_GRAPH_ATTRIBUTES'],
            dropout_rate=e.parameters['DROPOUT_RATE']
        )

        model.compile(
            loss=mse,
            metrics=mse,
            optimizer=ks.optimizers.Adam(learning_rate=e.parameters['LEARNING_RATE']),
            run_eagerly=False
        )

        return model



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['graph_attention_student',
 'graph_attention_student.examples',
 'graph_attention_student.experiments',
 'graph_attention_student.models']

package_data = \
{'': ['*'], 'graph_attention_student': ['templates/*']}

install_requires = \
['cairosvg>=2.5.2',
 'click>=7.1.2',
 'imageio>=2.19.0',
 'jinja2>=3.0.3',
 'kgcnn>=2.2.1',
 'matplotlib>=3.5.3',
 'nltk>=3.7',
 'numpy>=1.22.0',
 'orjson>=3.8.0',
 'poetry-bumpversion>=0.3.0',
 'pycomex>=0.8.4',
 'rdkit>=2022.9.1',
 'tensorflow>=2.8.0',
 'visual_graph_datasets>=0.7.2']

entry_points = \
{'console_scripts': ['graph_attention_student = '
                     'graph_attention_student.cli:cli']}

setup_kwargs = {
    'name': 'graph-attention-student',
    'version': '0.7.0',
    'description': 'MEGAN: Multi Explanation Graph Attention Network',
    'long_description': '|made-with-python| |made-with-kgcnn| |python-version| |os-linux|\n\n.. |os-linux| image:: https://img.shields.io/badge/os-linux-orange.svg\n   :target: https://www.python.org/\n\n.. |python-version| image:: https://img.shields.io/badge/Python-3.8.0-green.svg\n   :target: https://www.python.org/\n\n.. |made-with-kgcnn| image:: https://img.shields.io/badge/Made%20with-KGCNN-blue.svg\n   :target: https://github.com/aimat-lab/gcnn_keras\n\n.. |made-with-python| image:: https://img.shields.io/badge/Made%20with-Python-1f425f.svg\n   :target: https://www.python.org/\n\n.. image:: architecture.png\n    :width: 800\n    :alt: Architecture Overview\n\n================================================\nMEGAN: Multi Explanation Graph Attention Student\n================================================\n\nExplainable artificial intelligence (XAI) methods are expected to improve trust during human-AI interactions,\nprovide tools for model analysis and extend human understanding of complex problems. Attention-based models\nare an important subclass of XAI methods, partly due to their full differentiability and the potential to\nimprove explanations by means of explanation-supervised training. We propose the novel multi-explanation\ngraph attention network (MEGAN). Our graph regression and classification model features multiple explanation\nchannels, which can be chosen independently of the task specifications. We first validate our model on a\nsynthetic graph regression dataset, where our model produces single-channel explanations with quality\nsimilar to GNNExplainer. Furthermore, we demonstrate the advantages of multi-channel explanations on one\nsynthetic and two real-world datasets: The prediction of water solubility of molecular graphs and\nsentiment classification of movie reviews. We find that our model produces explanations consistent\nwith human intuition, opening the way to learning from our model in less well-understood tasks.\n\nInstallation\n============\n\nMain Installation\n-----------------\n\nClone the repository from github:\n\n.. code-block:: shell\n\n    git clone https://github.com/awa59kst120df/graph_attention_student.git\n\nThen in the main folder run a ``pip install``:\n\n.. code-block:: shell\n\n    cd graph_attention_student\n    pip3 install -e .\n\nAfterwards, you can check the install by invoking the CLI:\n\n.. code-block:: shell\n\n    python3 -m graph_attention_student.cli --version\n    python3 -m graph_attention_student.cli --help\n\nUsage\n=====\n\nComputational Experiments\n-------------------------\n\nIt is possible to list, show and execute all the computational experiments using a command line interface\nCLI.\n\n    *NOTE* Most of the experiments have a long runtime, ranging from ~2hrs to ~2days.\n    Furthermore, all of the experiments which do model training are currently configured to run on a GPU\n    and might crash if the GPU can either not be detected or does not have enough VRAM. This setting can\n    be changed in the corresponding experiment scripts\n\nAll the available experiments can be listed like this:\n\n.. code-block:: shell\n\n    python3 -m graph_attention_student.cli list\n\nThe details for a specific experiment can be viewed like this:\n\n.. code-block:: shell\n\n    python3 -m graph_attention_student.cli info [experiment_name]\n\nA new run of an experiment can be started like this. However, be aware that this might take some time.\n\n.. code-block::\n\n    python3 -m graph_attention_student.cli run [experiment_name]\n\nEach experiment will create a new archive folder, which will contain all the artifacts (such as visual\nexamples and the raw data) created during the runtime. The location of this archive folder can be found\nfrom the output generated by the experiment execution.\n\nArchived Experiments\n--------------------\n\nTo view the detailed data which was used in the making of the paper, go to\n``graph_attention_student/experiments``. The subfolders in that folder contain the archived experiments.\nThese contain extensive examples for each repetition of the various experiments as well as all of the raw\ndata collected during the execution of the experiments.\n\nMEGAN in code\n-------------\n\nThe MEGAN model is implemented as the ``MultiAttentionStudent`` class, which implements ``keras.Model``.\nThe implementation is based on the `kgcnn`_ library for graph convolutional networks for keras. For further\ninformation on loading graph structured data with `kgcnn`_ visit:\nhttps://github.com/aimat-lab/gcnn_keras\n\nThis is a simple example of how to use the model in the regression case:\n\n.. code-block:: python\n\n    import tensorflow as tf\n    import tensorflow.keras as ks\n    from graph_attention_student.training import NoLoss\n    from graph_attention_student.models import Megan\n\n    model = Megan(\n        # These lists define the number of layers and the number of hidden units in each layer for the\n        # various parts of the architecture\n        units=[9, 9, 9],  # The main convolutional layers\n        importance_units=[],  # The MLP that creates the node importances\n        final_units=[5, 1],  # The final MLP for graph embeddings\n        # Example for a regression problem. We need the prior knowledge about what range the values of the\n        # dataset will be expected to fall into...\n        regression_limits=(-3, +3),\n        # ... as well as a reference value.\n        regression_reference=0,\n        # This controls the weight of the explanation-only train step (gamma)\n        importance_factor=1.0,\n        importance_multiplier=5,\n        # This is the weight of the sparsity regularization\n        sparsity_factor=0.1,\n    )\n\n    # The model output is actually a three tuple: (prediction, node_importances, edge_importances).\n    # This allows the importances to be trained in a supervised fashion. If we don\'t want that,\n    # we can simply supply the NoLoss function instead.\n    model.compile(\n        loss=[ks.losses.MeanSquaredError(), NoLoss(), NoLoss()],\n        loss_weights=[1, 1, 1],\n        optimizer=ks.optimizers.Adam(0.001)\n    )\n\n    # model.fit() ...\n\n\n.. _kgcnn: https://github.com/aimat-lab/gcnn_keras\n.. _examples/solubility_regression.py: https://github.com/aimat-lab/graph_attention_student/tree/master/graph_attention_student/examples/solubility_regression.py\n.. _`GATv2`: https://github.com/tech-srl/how_attentive_are_gats\n\n---\n\nExamples\n========\n\nThe following examples show some of the *cherry picked* examples that show the explanatory capabilities of\nthe model.\n\nRB-Motifs Dataset\n-----------------\n\nThis is a synthetic dataset, which basically consists of randomly generated graphs with nodes of different\ncolors. Some of the graphs contain special sub-graph motifs, which are either blue-heavy or red-heavy\nstructures. The blue-heavy sub-graphs contribute a certain negative value to the overall value of the graph,\nwhile red-heavy structures contain a certain positive value.\n\nThis way, every graph has a certain value associated with it, which is between -3 and 3. The network was\ntrained to predict this value for each graph.\n\n.. image:: rb_motifs_example.png\n    :width: 800\n    :alt: Rb-Motifs Example\n\nThe examples shows from left to right: (1) The ground truth explanations, (2) a baseline MEGAN model trained\nonly on the prediction task, (3) explanation-supervised MEGAN model and (4) GNNExplainer explanations for a\nbasic GCN network. While the baseline MEGAN and GNNExplainer focus only on one of the ground truth motifs,\nthe explanation-supervised MEGAN model correctly finds both.\n\nWater Solubility Dataset\n------------------------\n\nThis is the `AqSolDB`_ dataset, which consists of ~10000 molecules and measured values for the solubility in\nwater (logS value).\n\nThe network was trained to predict the solubility value for each molecule.\n\n.. image:: solubility_example.png\n    :width: 800\n    :alt: Solubility Example.png\n\n.. _`AqSolDB`: https://www.nature.com/articles/s41597-019-0151-1\n\nMovie Reviews\n-------------\n\nOriginally the *MovieReviews* dataset is a natural language processing dataset from the `ERASER`_ benchmark.\nThe task is to classify the sentiment of ~2000 movie reviews collected from the IMDB database into the\nclasses "positive" and "negative". This dataset was converted into a graph dataset by considering all words\nas nodes of a graph and then connecting adjacent words by undirected edges with a sliding window of size 2.\nWords were converted into numeric feature vectors by using a pre-trained `GLOVE`_ model.\n\nExample for a positive review:\n\n.. image:: movie_reviews_pos.png\n    :width: 800\n    :alt: Positive Movie Review\n\nExample for a negative review:\n\n.. image:: movie_reviews_neg.png\n    :width: 800\n    :alt: Negative Movie Review\n\nExamples show the explanation channel for the "negative" class left and the "positive" class right.\nSentences with negative / positive adjectives are appropriately attributed to the corresponding channels.\n\n.. _`ERASER`: https://www.eraserbenchmark.com/\n.. _`GLOVE`: https://nlp.stanford.edu/projects/glove/\n\n',
    'author': 'xxx',
    'author_email': 'jonseb1998@gmail.com',
    'maintainer': 'xxx',
    'maintainer_email': 'jonseb1998@gmail.com',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.0,<3.11',
}


setup(**setup_kwargs)



import logging
import time
import typing as t

import tensorflow as tf
import tensorflow.keras as ks

from graph_attention_student.util import NULL_LOGGER


def gnnx_importances(model,
                     x,
                     y,
                     epochs: int = 100,
                     learning_rate: float = 0.01,
                     node_sparsity_factor: float = 1.0,
                     edge_sparsity_factor: float = 1.0,
                     model_kwargs: dict = {},
                     logger: logging.Logger = NULL_LOGGER,
                     log_step: int = 10):
    logger.info('creating explanations with gnn explainer')
    start_time = time.time()

    optimizer = ks.optimizers.Nadam(learning_rate=learning_rate)

    node_input = x[0]
    edge_input = x[1]
    edge_index_input = x[2]

    node_mask_ragged = tf.reduce_mean(tf.ones_like(node_input), axis=-1, keepdims=True)
    node_mask_variables = tf.Variable(node_mask_ragged.flat_values, trainable=True, dtype=tf.float64)

    edge_mask_ragged = tf.reduce_mean(tf.ones_like(edge_input), axis=-1, keepdims=True)
    edge_mask_variables = tf.Variable(edge_mask_ragged.flat_values, trainable=True, dtype=tf.float64)

    for epoch in range(epochs):

        with tf.GradientTape() as tape:
            node_mask = tf.RaggedTensor.from_nested_row_splits(
                node_mask_variables,
                nested_row_splits=node_mask_ragged.nested_row_splits
            )
            node_masked = node_input * node_mask

            edge_mask = tf.RaggedTensor.from_nested_row_splits(
                edge_mask_variables,
                nested_row_splits=edge_mask_ragged.nested_row_splits
            )
            edge_masked = edge_input * edge_mask

            out = model([
                node_masked,
                edge_masked,
                edge_index_input,
                *x[3:]
            ], **model_kwargs)

            loss = tf.cast(tf.reduce_mean(tf.square(y - out)), dtype=tf.float64)
            loss += node_sparsity_factor * tf.reduce_mean(tf.abs(node_mask))
            loss += edge_sparsity_factor * tf.reduce_mean(tf.abs(edge_mask))

        trainable_vars = [node_mask_variables, edge_mask_variables]
        gradients = tape.gradient(loss, trainable_vars)
        optimizer.apply_gradients(zip(gradients, trainable_vars))

        if epoch % log_step == 0:
            logger.info(f' ({epoch}/{epochs})'
                        f' - loss: {loss}'
                        f' - elapsed time: {time.time() - start_time:.2f}')

    return (
        node_mask,
        edge_mask
    )


import typing as t

import tensorflow as tf
import tensorflow.keras as ks
from kgcnn.layers.conv.gcn_conv import GCN
from kgcnn.layers.modules import DenseEmbedding
from kgcnn.layers.pooling import PoolingNodes
from kgcnn.literature.GNNExplain import GNNInterface


class AbstractGradientModel(ks.models.Model):
    """
    This is an abstract base class for models which will provide their gradient information for further
    processing. This means that during the inference call, these kinds of models have to maintain their
    own gradient tape to record the full inference process and then use this to provide gradients of the
    input and intermediate node / edge embeddings with respect to the output.
    """

    def call_with_gradients(self, inputs, training=True):
        """
        **How the list of gradients should look like**

        Assuming output dimension H, number of nodes V, node features N and batch size B, as well as the
        number of convolutional layers L in the network.

        The list should consist of L+1 elements, where the first element is always the gradients with respect
        to the input and after that the gradients of the intermediate node/edge embeddings.

        Each element should be a ragged tensor with the dimension ([B], [V], N, H)

        :return: This has to be a tuple of four values:
            - The main output prediction "y" of the model
            - A list containing the gradients of the input as well as the intermediate node(!) embeddings
              w.r.t to the output "y".
            - A list containing the gradients of the input as well as the intermediate edge(!) embeddings
              w.r.t to the output "y".
            - The gradient tape used to create the previously mentioned gradients
        """
        raise NotImplementedError()

    def call(self, inputs, training=True, *args, **kwargs):
        y, node_gradients, edge_gradients, tape = self.call_with_gradients(inputs, training, *args, **kwargs)
        return y

    def calculate_gradient_info(self,
                                outs: t.List[t.List[tf.Tensor]],
                                activations: t.List[tf.Tensor],
                                tape: tf.GradientTape,
                                batch_size):
        """
        This method is a boilerplate implementation that will calculate the gradient info list, given a list
        of model outputs, a list of layer activation tensors, the gradient tape and the batch size.

        Assuming the following variables:
        - B: batch dimension
        - O: output dimension
        - L: number of convolutional layers / number of intermediate embeddings w.r.t which the gradients
            are to be calculated
        - V: Node or edge shape
        - F: Feature shape

        :param outs: This is supposed to be a list of lists of zero dimensional tensors. Generally, the
            output of a regression / classification network will be a vector of shape ([B], O). This list
            has to basically be this tensor only in list format. This list should be created by applying
            the corresponding indexing slices to the output vector to split it into the nested format of
            zero dimensional tensor objects. Note that this indexing operation has to be done INSIDE the
            gradient tape context, otherwise the calculation of the gradients will not work!
        :param activations: This is a list with L elements, where each element is an intermediate node /
            edge embedding vector from the call process of the network. In this method one gradient vector
            will be created for each of the outputs w.r.t to each of these embeddings. The resulting
            gradient tensors will have the same shape as the embeddings!
        :param tape: The GradientTape which was used during the call process.
        :param batch_size: The batch size B that was used to obtain the outputs and the embeddings. This has
            to be exactly correct!
        :return: A list of tuples with the length L. Each tuple contains two elements. The first one is the
            exact same embedding tensor that was passed to the method in the "activations" parameter.
            This tensor has the shape ([B], [V], F). The second element is the corresponding gradient
            tensor that was created for that very embedding with the shape ([B], [V], F, O). Note that the
            gradient has an additional dimension for the output shape. This is because gradients are
            created for every output.
        """
        gradient_info = []
        for activation in activations:
            gradients = []
            for b in range(batch_size):

                # In this inner loop here we manage the fact that the network may have multiple outputs
                # O != 1, so we iterate over every one of the outputs for the current element of the batch
                # add them to "gradient_stack" and then concat them in the end. The important part here is
                # that we want to stack them such that the resulting shape is ([B], [V], O, N) and not
                # ([B], [V], N, O) as you would intuitively do it by concat over the last dimension!
                gradient_stack = []
                for out in outs[b]:
                    # What we get out of this operation is the following shape ([B], [V], N) but technically
                    # the first dimension is pointless, because we are calculating the gradients of only one
                    # specific output value in the batch with respect to every element in the batch. Only
                    # one element in this dimension will be != zero. But actually we need to do it this way
                    # so that the ragged concat operation further down the line will work. If we already only
                    # select the appropriate element here, we get a serious issue with the stacking of ragged
                    # dimensions later.
                    gradient = tape.gradient(out, activation)
                    gradient = tf.expand_dims(gradient, axis=-2)
                    gradient_stack.append(gradient)

                # gradient_stack will be a vector of shape ([B], [V], O, N, 1). The last dimension we have
                # expanded here because in the following we need to stack this over the batch dimension
                # again.
                gradient_stack = tf.concat(gradient_stack, axis=-2)
                gradient_stack = tf.expand_dims(gradient_stack, axis=-1)

                gradients.append(gradient_stack)

            # results in a tensor of shape ([B], [V], O, N, [B])
            gradients = tf.concat(gradients, axis=-1)
            # So this is the point AFTER the concat operation where we can savely get rid of the one
            # batch dimension we have uselessly kept before.
            # results in fully populated gradient vector of shape ([B], [V], N) exactly as we wanted
            gradients = tf.reduce_sum(gradients, axis=-1)
            gradient_info.append((activation, gradients))

        return gradient_info


class GcnGradientModel(AbstractGradientModel, GNNInterface):

    def __init__(self,
                 batch_size: int,
                 units: t.List[int],
                 final_units: t.List[int],
                 layer_cb: t.Callable = lambda units: GCN(units=units),
                 pooling_method: str = 'mean',
                 final_activation: str = 'linear'):
        AbstractGradientModel.__init__(self)
        GNNInterface.__init__(self)

        self.batch_size = batch_size
        self.units = units
        self.num_outputs = final_units[-1]

        self.conv_layers = []
        for k in self.units:
            lay = layer_cb(k)
            self.conv_layers.append(lay)

        self.lay_pooling = PoolingNodes(pooling_method=pooling_method)

        self.final_units = final_units
        self.final_acts = ['kgcnn>leaky_relu' for _ in final_units]
        self.final_acts[-1] = final_activation
        self.final_layers = []
        for k, act in zip(self.final_units, self.final_acts):
            lay = DenseEmbedding(units=k, activation=act)
            self.final_layers.append(lay)

    def call_with_gradients(self, inputs, training=True, create_gradients=True, batch_size=None):
        if batch_size is None:
            batch_size = self.batch_size

        node_input, edge_input, edge_index_input = inputs
        x = node_input

        with tf.GradientTape(persistent=True) as tape:
            tape.watch(node_input)
            tape.watch(edge_input)

            edge_activations = [edge_input]
            node_activations = [x]
            for lay in self.conv_layers:
                x = lay([x, edge_input, edge_index_input])
                node_activations.append(x)

            out = self.lay_pooling(x)

            for lay in self.final_layers:
                out = lay(out)

            outs = [out[b, :] for b in range(batch_size)]
            outs_multi = [[out[b, o] for o in range(self.num_outputs)] for b in range(batch_size)]

        if create_gradients:
            edge_gradient_info = self.calculate_gradient_info(outs_multi, edge_activations, tape, batch_size)
            node_gradient_info = self.calculate_gradient_info(outs_multi, node_activations, tape, batch_size)
        else:
            node_gradient_info = []
            edge_gradient_info = []

        return out, node_gradient_info, edge_gradient_info, tape

    def call(self,
             inputs,
             training=True,
             batch_size=None,
             create_gradients=True,
             return_gradients=False):
        y, node_info, edge_info, tape = self.call_with_gradients(
            inputs,
            training=training,
            batch_size=batch_size,
            create_gradients=create_gradients
        )

        if return_gradients:
            return y, node_info, edge_info
        else:
            return y


def grad_importances(gradient_info: t.List[t.Tuple[tf.Tensor, tf.Tensor]],
                     use_relu: bool = False,
                     use_absolute: bool = False,
                     keepdims: bool = False):
    # shapes as we get them:
    # gradients: ([B], [V], O, F)
    # activation: ([B], [V], F)

    activation, gradients = gradient_info[0]
    importances = tf.reduce_sum(
        # We need to expand the activations here in the second last dimension to match the additional
        # output shape O of the gradients.
        gradients * tf.expand_dims(activation, axis=-2),
        axis=-1,
        keepdims=keepdims
    )
    if use_relu:
        importances = ks.backend.relu(importances)
    if use_absolute:
        importances = tf.abs(importances)

    return importances


def grad_cam_importances(gradient_info: t.List[t.Tuple[tf.Tensor, tf.Tensor]],
                         use_average: bool = False,
                         use_relu: bool = False,
                         use_absolute: bool = False,
                         averaging_depth: int = 2,
                         keepdims: bool = False):
    importances = []
    for activation, gradients in gradient_info:
        # shapes as we get them:
        # gradients: ([B], [V], O, F)
        # activation: ([B], [V], F)

        alpha = tf.reduce_mean(gradients, axis=1, keepdims=True)
        local_importances = tf.reduce_sum(
            # We need to expand the activations here in the second last dimension to match the additional
            # output shape O of the gradients.
            alpha * tf.expand_dims(activation, axis=-2),
            axis=-1,
            keepdims=keepdims
        )

        if use_relu:
            local_importances = ks.backend.relu(local_importances)
        if use_absolute:
            local_importances = tf.abs(local_importances)

        importances.append(local_importances)

    if use_average:
        importances = tf.concat(importances[-averaging_depth:], axis=-1)
    else:
        importances = importances[-1]

    importances = tf.reduce_mean(importances, axis=-1, keepdims=True)
    return importances



import os
import pathlib
import typing as t

import tensorflow as tf
import tensorflow.keras as ks
from pycomex.util import Skippable
from pycomex.experiment import SubExperiment
from kgcnn.layers.modules import DenseEmbedding, DropoutEmbedding, LazyConcatenate, LazyAverage
from kgcnn.layers.conv.gat_conv import AttentionHeadGATV2
from kgcnn.layers.pooling import PoolingGlobalEdges

from graph_attention_student.training import mse

# == DATASET PARAMETERS ==
USE_EDGE_ATTRIBUTES = True


# == MODEL PARAMETERS ==
MODEL_NAME = 'GATv2'

class Gatv2Model(ks.models.Model):

    def __init__(self,
                 units: t.List[int],
                 final_units: t.List[int],
                 pooling_method: str = 'sum',
                 num_heads: int = 2,
                 activation: str = 'kgcnn>leaky_relu',
                 final_pooling: str = 'sum',
                 final_activation: str = 'linear',
                 dropout_rate: float = 0.0,
                 use_edge_features: bool = True):
        super(Gatv2Model, self).__init__()

        self.lay_dropout = DropoutEmbedding(rate=dropout_rate)
        self.conv_layers = []
        for k in units:
            heads = []
            for _ in range(num_heads):
                lay = AttentionHeadGATV2(
                    units=k,
                    activation=activation,
                    use_edge_features=use_edge_features,
                )
                heads.append(lay)

            self.conv_layers.append(heads)

        self.lay_combine = LazyAverage()
        self.lay_combine = LazyConcatenate()
        self.lay_pooling = PoolingGlobalEdges(pooling_method=final_pooling)
        self.lay_concat = LazyConcatenate(axis=-1)

        self.final_layers = []
        self.final_activations = ['relu' for _ in final_units]
        self.final_activations[-1] = final_activation
        for k, act in zip(final_units, self.final_activations):
            lay = DenseEmbedding(units=k, activation=act)
            self.final_layers.append(lay)

    def call(self, inputs, training=False):
        node_input, edge_input, edge_index_input = inputs

        x = node_input
        for heads in self.conv_layers:

            xs = []
            for lay in heads:
                _x = lay([x, edge_input, edge_index_input])
                xs.append(_x)

            x = self.lay_combine(xs)
            if training:
                x = self.lay_dropout(x)

        final = self.lay_pooling(x)
        for lay in self.final_layers:
            final = lay(final)

        return final


MODEL_NAME = 'GATv2'
UNITS: t.List[int] = [32, 32, 32]
DROPOUT_RATE: float = 0.0
# The number of attention heads to be used in the network.
NUM_HEADS: int = 2

# == TRAINING PARAMETERS ==
BATCH_SIZE = 256
REPETITIONS = 5

# == EXPERIMENT PARAMETERS ==
PATH = pathlib.Path(__file__).parent.absolute()
EXPERIMENT_PATH = os.path.join(PATH, 'vgd_single_gnnx_aqsoldb.py')
BASE_PATH = os.getcwd()
NAMESPACE = 'results/vgd_single_gnnx_aqsoldb_gatv2'
DEBUG = True
with Skippable(), (se := SubExperiment(EXPERIMENT_PATH, BASE_PATH, NAMESPACE, globals())):

    @se.hook('create_model', replace=True)
    def create_model(e):
        e.info('GATv2 Model')
        model = Gatv2Model(
            units=e.parameters['UNITS'],
            final_units=e.parameters['FINAL_UNITS'],
            dropout_rate=e.parameters['DROPOUT_RATE'],
        )
        model.compile(
            loss=mse,
            metrics=mse,
            optimizer=e.parameters['OPTIMIZER_CB'](),
            run_eagerly=False
        )
        return model



"""
This example shows how to train a KGCNN model for multitask graph regression application based on an
existing "visual graph dataset".

This module loads a multitask regression visual graph dataset and trains a multi-layered GNN model to
predict all the targets simultaneously. This base module uses standard GCN layers and a global pooling
operation followed by a MLP to make the prediction.
The module calculates the most important regression performance metrics for each of the targets, such as
MSE and R2 and plots these results for each repetition of the training process.

This module implements various hooks, which means it can be used as the basis for various modifications,
such as changing the used model, by sub-experiment implementation.

**CHANGELOG**

0.1.0 - 17.01.2023 - Initial version

0.2.0 - 18.01.2023 - The experiment can now be repeated multiple independent times with different train
test splits of the dataset.
Analysis now renders a latex table containing the combined results of all the different targets
"""
# standard library
import os
import warnings
warnings.filterwarnings("ignore")
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import random
import typing as t

# third party
import numpy as np
import matplotlib.pyplot as plt
import tensorflow as tf
import tensorflow.keras as ks
from matplotlib.backends.backend_pdf import PdfPages
from sklearn.metrics import r2_score
from sklearn.metrics import mean_squared_error
from pycomex.experiment import Experiment
from pycomex.util import Skippable
from visual_graph_datasets.data import load_visual_graph_dataset
from kgcnn.layers.conv.gcn_conv import GCN
from kgcnn.layers.pooling import PoolingGlobalEdges
from kgcnn.layers.modules import DenseEmbedding, DropoutEmbedding, LazyConcatenate

# local
import graph_attention_student.typing as tc
from graph_attention_student.data import process_graph_dataset
from graph_attention_student.training import mse
from graph_attention_student.training import LogProgressCallback
from graph_attention_student.visualization import plot_regression_fit
from graph_attention_student.util import latex_table, latex_table_element_mean
from graph_attention_student.util import render_latex

VERSION = '0.1.0'

# == DATASET PARAMETERS ==
# These parameters are used to specify the dataset to be used for the training as well as additional
# properties of the dataset such as the train test split for example.

# The name of the visual graph dataset to use for this experiment.
VISUAL_GRAPH_DATASET_PATH = os.path.expanduser('~/.visual_graph_datasets/datasets/organic_solvents')
# The ratio of how many elements of the dataset are supposed to be used for the training dataset.
# The rest of them will be used for the test set.
TRAIN_RATIO = 0.8
# The number of target values for each graph in the dataset.
NUM_TARGETS = 4
# Whether the dataset already contains importance (explanation) ground truth annotations.
# most of the time this will most likely not be the case
HAS_IMPORTANCES: bool = False
# IF the dataset includes global graph attributes and if they are supposed to be used in the training
# process, this flag has to be set to True.
USE_GRAPH_ATTRIBUTES: bool = True
# The ratio of the test set to be used as examples for the visualization of the explanations
EXAMPLES_RATIO: float = 0.2
# The string names of the target values in the order in which they appear in the dataset as well
# which will be used in the labels for the result visualizations
TARGET_NAMES: t.List[str] = [
    'water',
    'benzene',
    'acetone',
    'ethanol',
]


# == MODEL PARAMETERS ==

class GcnModel(ks.models.Model):

    def __init__(self,
                 units: t.List[int],
                 final_units: t.List[int],
                 pooling_method: str = 'sum',
                 activation: str = 'kgcnn>leaky_relu',
                 final_pooling: str = 'sum',
                 final_activation: str = 'linear',
                 dropout_rate: float = 0.0,
                 use_graph_attributes: bool = False):
        super(GcnModel, self).__init__()
        self.use_graph_attributes = use_graph_attributes

        self.lay_dropout = DropoutEmbedding(rate=dropout_rate)
        self.conv_layers = []
        for k in units:
            lay = GCN(units=k, pooling_method=pooling_method, activation=activation)
            self.conv_layers.append(lay)

        self.lay_pooling = PoolingGlobalEdges(pooling_method=final_pooling)
        self.lay_concat = LazyConcatenate(axis=-1)

        self.final_layers = []
        self.final_activations = ['relu' for _ in final_units]
        self.final_activations[-1] = final_activation
        for k, act in zip(final_units, self.final_activations):
            lay = DenseEmbedding(units=k, activation=act)
            self.final_layers.append(lay)

    def call(self, inputs, training=False):
        if self.use_graph_attributes:
            node_input, edge_input, edge_index_input, graph_input = inputs
        else:
            node_input, edge_input, edge_index_input = inputs

        x = node_input
        for lay in self.conv_layers:
            x = lay([x, edge_input, edge_index_input])
            if training:
                x = self.lay_dropout(x)

        final = self.lay_pooling(x)
        if self.use_graph_attributes:
            final = self.lay_concat([final, graph_input])

        for lay in self.final_layers:
            final = lay(final)

        return final


MODEL_NAME = 'GCN'
UNITS = [64, 64, 64]
FINAL_UNITS = [30, 10]
DROPOUT_RATE = 0.2

# == TRAINING PARAMETERS ==
# These parameters are for the learning / training process of the model.

# The training can be repeated multiple times to get a statistical measure for the performance of the model.
# This determines how many repetitions are made
REPETITIONS: int = 5
# This string defines which device is used for the training process. On a normal setup for a normal PC the
# main two options would be 'cpu:0' and 'gpu:0' which uses the CPU and GPU respectively. Training on a
# GPU is usually more efficient, but one might encounter errors in some cases such as tensor operations
# which are not possible on GPU or insufficient video memory for large batches. In such a case one should
# switch to CPU instead, which will be slower.
DEVICE: str = 'gpu:0'
# This is the learning rate of the optimization process used by the Optimizer.
LEARNING_RATE: float = 0.01
# The batch size is the number of elements from the training dataset which will be concatenated into a
# single tensor, which will then be used for one forward & backward pass of the model, ultimately resulting
# in one weight update.
BATCH_SIZE: int = 256
# How many times the entire training dataset will be iterated over during the training process
EPOCHS: int = 1000
# This controls how often the loss and the metrics for the training process will be printed to the
# console.
EPOCH_STEP: int = 50

# == EVALUATION PARAMETERS ==
# These parameters determine the behavior and the look of the evaluation process, which includes for example
# the generation of visualizations.
LOG_STEP_EVAL: int = 1000

# == EXPERIMENT PARAMETERS ==
BASE_PATH = os.getcwd()
NAMESPACE = 'results/vgd_multitask_gnn'
DEBUG = True
with Skippable(), (e := Experiment(BASE_PATH, NAMESPACE, globals())):
    FINAL_UNITS += [NUM_TARGETS]

    e.info('starting experiment...')

    e.info('loading dataset...')
    name_data_map, index_data_map = load_visual_graph_dataset(
        path=VISUAL_GRAPH_DATASET_PATH,
        logger=e.logger,
        log_step=LOG_STEP_EVAL,
        metadata_contains_index=True
    )
    dataset_indices = list(sorted(index_data_map.keys()))
    dataset_length = len(index_data_map)

    # This callback can be used to apply modifications to each individual graph of the VGD. In this default
    # implementation the edge attributes are replaced with constant edge weights, since the GCN layers are
    # only able to deal with edge weights and not edge attribute vectors.
    @e.hook('modify_g')
    def modify_g(_e, g):
        g['edge_attributes'] = np.array([[1.0] for _ in g['edge_indices']])
        return g

    dataset: t.List[tc.GraphDict] = []
    for index in dataset_indices:
        data = index_data_map[index]
        g = data['metadata']['graph']

        g = e.apply_hook('modify_g', g=g)

        dataset.append(g)

    e.info(f'loaded dataset with {len(dataset)} elements')

    # -- TRAINING THE MODEL --

    @e.hook('create_model', default=True)
    def create_model(_e):
        model = GcnModel(
            units=_e.parameters['UNITS'],
            final_units=_e.parameters['FINAL_UNITS'],
            use_graph_attributes=_e.parameters['USE_GRAPH_ATTRIBUTES'],
            dropout_rate=_e.parameters['DROPOUT_RATE'],
        )
        model.compile(
            loss=mse,
            metrics=mse,
            optimizer=ks.optimizers.Adam(learning_rate=_e.parameters['LEARNING_RATE']),
            run_eagerly=False
        )

        return model


    @e.hook('model_training', default=True)
    def model_training(_e, x_train, y_train, x_test, y_test):
        history = model.fit(
            x_train,
            y_train,
            batch_size=_e.parameters['BATCH_SIZE'],
            epochs=_e.parameters['EPOCHS'],
            validation_data=(x_test, y_test),
            validation_freq=1,
            callbacks=LogProgressCallback(
                logger=_e.logger,
                epoch_step=_e.parameters['EPOCH_STEP'],
                identifier=f'val_mean_squared_error'
            ),
            verbose=[]
        )

        return history


    for rep in range(REPETITIONS):
        e.info(f'REPETITION ({rep+1}/{REPETITIONS})')

        e.info('creating train test split...')
        train_indices = random.sample(dataset_indices, k=int(TRAIN_RATIO * len(dataset_indices)))
        test_indices = [index for index in dataset_indices if index not in train_indices]
        e[f'train_indices/{rep}'] = train_indices
        e[f'test_indices/{rep}'] = test_indices
        e.info(f'determined {len(train_indices)} train_indices and {len(test_indices)} test indices')

        # This turns the list of graph dicts into the final form which we need for the training of the model:
        # keras RaggedTensors which contain all the graphs.
        x_train, y_train, x_test, y_test = process_graph_dataset(
            dataset,
            train_indices=train_indices,
            test_indices=test_indices,
            use_importances=False,
            use_graph_attributes=e.parameters['USE_GRAPH_ATTRIBUTES'],
        )
        e.info(f'num elements x tuple: {len(x_train)}')
        e.info(f'num elements y tuple: {len(y_train)}')

        with tf.device(DEVICE):
            model = e.apply_hook('create_model')

            e.info('starting model training...')
            e.apply_hook('before_training')

            history = e.apply_hook(
                'model_training',
                x_train=x_train,
                y_train=y_train,
                x_test=x_test,
                y_test=y_test
            )
            e[f'history/{rep}'] = history

            # -- EVALUATING ON TEST SET --
            e.info('evaluating on test set...')
            e.apply_hook('before_evaluation')

            out_pred = model(x_test)
            out_pred = np.array(out_pred.numpy())
            out_true = y_test[0]

            # ~ visualizing the evaluation results
            # This section generates a PDF which contains the R2 scores for all the different target values,
            # calculated using the test set predictions.
            # The PDF also visualizes the regression results using a regression "scatter" plot.
            pdf_path = os.path.join(e.path, f'training_{rep:02d}.pdf')
            with PdfPages(pdf_path) as pdf:
                n_cols = len(TARGET_NAMES)
                n_rows = 1

                # First page is the regressions scatter plots for each of the targets
                fig, rows = plt.subplots(ncols=n_cols, nrows=n_rows, figsize=(10 * n_cols, 10), squeeze=False)
                for i, name in enumerate(TARGET_NAMES):
                    ax = rows[0][i]

                    values_true = []
                    values_pred = []
                    for value_true, value_pred in zip(out_true, out_pred):
                        if not np.isnan(value_true[i]) and not np.isnan(value_pred[i]):
                            values_true.append(value_true[i])
                            values_pred.append(value_pred[i])

                    plot_regression_fit(
                        values_true=values_true,
                        values_pred=values_pred,
                        ax=ax
                    )

                    r2_value = r2_score(values_true, values_pred)
                    mse_value = mean_squared_error(values_true, values_pred)
                    rmse_value = np.sqrt(mse_value)
                    ax.set_title(f'"{name}"\n'
                                 f'r2: {r2_value:.2f} \n'
                                 f'mse: {mse_value:.2f} \n'
                                 f'rmse: {rmse_value:.2f}')

                    # We also want to save it to the experiment store so that we can use that information
                    # again later on.
                    e[f'r2/{name}/{rep}'] = float(r2_value)
                    e[f'mse/{name}/{rep}'] = float(mse_value)
                    e[f'rmse/{name}/{rep}'] = float(rmse_value)

                pdf.savefig(fig)
                plt.close(fig)


with Skippable(), e.analysis:

    # Creating latex code to display the results in a table
    e.info('rendering latex table...')
    column_names = [
        r'Target Name',
        r'$\text{MSE} \downarrow $',
        r'$\text{RMSE} \downarrow $',
        r'$R^2 \uparrow $',
    ]
    rows = []
    for name in TARGET_NAMES:
        row = []

        row.append(name)
        row.append([e[f'mse/{name}/{rep}'] for rep in range(REPETITIONS)])
        row.append([e[f'rmse/{name}/{rep}'] for rep in range(REPETITIONS)])
        row.append([e[f'r2/{name}/{rep}'] for rep in range(REPETITIONS)])

        rows.append(row)

    content, table = latex_table(
        column_names=column_names,
        rows=rows,
        list_element_cb=latex_table_element_mean,
        caption=f'Results of {REPETITIONS} repetition(s) of ' + r'\textbf{' + MODEL_NAME + '}'
    )
    e.commit_raw('table.tex', table)
    pdf_path = os.path.join(e.path, 'table.pdf')
    render_latex({'content': table}, output_path=pdf_path)
    e.info('rendered latex table')



"""
Inherits ""

**CHANGLELOG**

0.1.0 - xx.xx.2023 - Initial version
"""
import os
import pathlib
import typing as t

import tensorflow as tf
import tensorflow.keras as ks
import numpy as np
from pycomex.util import Skippable
from pycomex.experiment import SubExperiment
from kgcnn.data.utils import ragged_tensor_from_nested_numpy
from kgcnn.literature.PAiNN import make_model

from graph_attention_student.training import mse
from graph_attention_student.training import LogProgressCallback

# == MODEL PARAMETERS ==
MODEL_NAME = 'PAiNN'
EQUIV_INITIALIZE_ARGS = {'dim': 3, 'method': 'zeros'}
BESSEL_BASIS = {'num_radial': 20, 'cutoff': 5.0, 'envelope_exponent': 5}
CONV_ARGS = {'units': 36, 'cutoff': None, 'conv_pool': 'sum'}
UPDATE_ARGS = {'units': 36}
DEPTH = 3
OUTPUT_MLP = {
    'units': [32, 16, 1],
    'use_bias': [True, True, True],
    'activation': ['relu', 'relu', 'linear']
}

# == TRAINING PARAMETERS ==
BATCH_SIZE = 32
OPTIMIZER_CB = lambda: ks.optimizers.Nadam(learning_rate=0.001)

# == EXPERIMENT PARAMETERS ==
PATH = pathlib.Path(__file__).parent.absolute()
EXPERIMENT_PATH = os.path.join(PATH, 'vgd_single_gnnx_aqsoldb.py')
BASE_PATH = PATH
NAMESPACE = 'results/vgd_single_gnnx_aqsoldb_painn'
with Skippable(), (se := SubExperiment(EXPERIMENT_PATH, BASE_PATH, NAMESPACE, globals())):

    @se.hook('process_dataset')
    def process_dataset(e, dataset, train_indices, test_indices):
        node_attributes_train = [dataset[i]['node_attributes'] for i in train_indices]
        node_attributes_test = [dataset[i]['node_attributes'] for i in test_indices]
        e['num_attributes'] = dataset[0]['node_attributes'].shape[1]

        node_coordinates_train = [dataset[i]['node_coordinates'] for i in train_indices]
        node_coordinates_test = [dataset[i]['node_coordinates'] for i in test_indices]

        edge_indices_train = [dataset[i]['edge_indices'] for i in train_indices]
        edge_indices_test = [dataset[i]['edge_indices'] for i in test_indices]

        graph_labels_train = [dataset[i]['graph_labels'] for i in train_indices]
        graph_labels_test = [dataset[i]['graph_labels'] for i in test_indices]

        node_importances_train = [dataset[i]['node_importances'] for i in train_indices]
        node_importances_test = [dataset[i]['node_importances'] for i in test_indices]

        edge_importances_train = [dataset[i]['edge_importances'] for i in train_indices]
        edge_importances_test = [dataset[i]['edge_importances'] for i in test_indices]

        x_train = (
            ragged_tensor_from_nested_numpy(node_attributes_train),
            ragged_tensor_from_nested_numpy(node_coordinates_train),
            ragged_tensor_from_nested_numpy(edge_indices_train)
        )

        x_test = (
            ragged_tensor_from_nested_numpy(node_attributes_test),
            ragged_tensor_from_nested_numpy(node_coordinates_test),
            ragged_tensor_from_nested_numpy(edge_indices_test)
        )

        y_train = (
            np.array(graph_labels_train),
            ragged_tensor_from_nested_numpy(node_importances_train),
            ragged_tensor_from_nested_numpy(edge_importances_train),
        )

        y_test = (
            np.array(graph_labels_test),
            ragged_tensor_from_nested_numpy(node_importances_test),
            ragged_tensor_from_nested_numpy(edge_importances_test),
        )

        return x_train, y_train, x_test, y_test

    @se.hook('create_model')
    def create_model(e):
        model = make_model(
            inputs=[
                {'shape': (None, e['num_attributes']), 'name': 'node_attributes', 'dtype': 'float32', 'ragged': True},
                {'shape': (None, 3), 'name': 'node_coordinates', 'dtype': 'float32', 'ragged': True},
                {'shape': (None, 2), 'name': 'edge_indices', 'dtype': 'int64', 'ragged': True},
            ],
            input_embedding={'node': {'input_dim': e['num_attributes'], 'output_dim': 32}},
            equiv_initialize_kwargs=e.parameters['EQUIV_INITIALIZE_ARGS'],
            bessel_basis=e.parameters['BESSEL_BASIS'],
            conv_args=e.parameters['CONV_ARGS'],
            update_args=e.parameters['UPDATE_ARGS'],
            depth=e.parameters['DEPTH'],
            output_mlp=e.parameters['OUTPUT_MLP'],
            node_normalization=False,
            equiv_normalization=False,
            output_embedding='graph',
            output_to_tensor=False,
        )
        model.compile(
            loss=mse,
            metrics=mse,
            optimizer=e.parameters['OPTIMIZER_CB'](),
            run_eagerly=False
        )
        return model

    @se.hook('fit_model')
    def fit_model(e, model, x_train, y_train, x_test, y_test):
        history = model.fit(
            x_train,
            y_train,
            epochs=e.parameters['EPOCHS'],
            batch_size=e.parameters['BATCH_SIZE'],
            validation_data=(x_test, y_test),
            validation_freq=1,
            callbacks=[
                LogProgressCallback(
                    e.logger,
                    epoch_step=5,
                    identifier='val_mean_squared_error'
                )
            ],
            verbose=0
        )
        return history.history



"""
This example shows how the internal implementation of GnnExplainer can be used to generate explanations
for the RbMotifs dataset
"""
import os
import warnings
warnings.filterwarnings("ignore")
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
import random
import typing as t

import tensorflow as tf
import tensorflow.keras as ks
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from pycomex.experiment import Experiment
from pycomex.util import Skippable
from sklearn.metrics import r2_score
from sklearn.metrics import roc_auc_score
from imageio.v2 import imread
from kgcnn.layers.conv.gcn_conv import GCN
from kgcnn.layers.modules import DenseEmbedding
from kgcnn.layers.pooling import PoolingNodes

import graph_attention_student.typing as tc
from graph_attention_student.util import DATASETS_FOLDER
from graph_attention_student.util import array_normalize
from graph_attention_student.data import process_graph_dataset
from graph_attention_student.data import load_eye_tracking_dataset_dict
from graph_attention_student.training import LogProgressCallback
from graph_attention_student.training import NoLoss, ExplanationLoss
from graph_attention_student.training import mae
from graph_attention_student.models import gnnx_importances
from graph_attention_student.visualization import plot_node_importances, plot_edge_importances


# == DATASET PARAMETERS ==
DATASET_PATH = os.path.join(DATASETS_FOLDER, 'rb_dual_motifs')
METADATA_CONTAINS_INDICES = True
TRAIN_RATIO = 0.8
NUM_EXAMPLES = 100

# == MODEL PARAMETERS ==
# We only need the most simple of GNN's for the demonstration
class Model(ks.models.Model):

    def __init__(self):
        super(Model, self).__init__()
        self.conv_layers = [GCN(units=k) for k in [5, 5, 5]]
        self.lay_pooling = PoolingNodes(pooling_method='sum')
        self.lay_dense = DenseEmbedding(units=1)

    def call(self, inputs):
        node_input, edge_input, edge_index_input = inputs
        x = node_input
        for lay in self.conv_layers:
            x = lay([x, edge_input, edge_index_input])
        out = self.lay_pooling(x)
        out = self.lay_dense(out)
        return out

# == TRAINING PARAMETERS ==
EPOCHS = 100
BATCH_SIZE = 32
LEARNING_RATE = 0.01

# == EVALUATION PARAMETERS ==
EVAL_BATCH_SIZE = 600

# == EXPERIMENT PARAMETERS ==
PATH = os.getcwd()
NAMESPACE = 'results/gnnx_example'
DEBUG = True
with Skippable(), (e := Experiment(base_path=PATH, namespace=NAMESPACE, glob=globals())):
    e.info('starting gnes example...')

    e.info('loading rb motifs dataset...')
    dataset_dict = load_eye_tracking_dataset_dict(DATASET_PATH, logger=e.logger)
    dataset_size = len(dataset_dict)
    dataset_index_dict = {}
    indices = []
    dataset: t.List[tc.GraphDict] = []
    for c, (name, data) in enumerate(dataset_dict.items()):
        metadata = data['metadata']
        g = metadata['graph']

        if METADATA_CONTAINS_INDICES:
            index = metadata['index']
        else:
            index = c

        g['node_importances'] = np.sum(np.array(g['multi_node_importances']), axis=-1, keepdims=True)
        g['edge_importances'] = np.sum(np.array(g['multi_edge_importances']), axis=-1, keepdims=True)

        g['edge_attributes'] = np.array(g['edge_attributes'], dtype=np.float)

        # With this here we want to support the possibility that the "value" annotation might already
        # be a multi target annotation (as is the case with classification datasets for example) but usually
        # this will be a single regression value and we will have to wrap that into a list first
        if isinstance(metadata['value'], list):
            g['graph_labels'] = np.array(metadata['value'])
        else:
            g['graph_labels'] = np.array([metadata['value']])

        indices.append(index)
        dataset_index_dict[index] = data

    dataset = [t[1]['metadata']['graph'] for t in sorted(dataset_index_dict.items(), key=lambda t: t[0])]
    e.info('determining train test split...')
    # At this point we need to determine the number of training samples exactly such that it is divisible
    # by the batch size! Because the implementation of GradCAM currently requires that
    num_train_samples = int(TRAIN_RATIO * dataset_size)
    if num_train_samples % BATCH_SIZE != 0:
        remaining = num_train_samples % BATCH_SIZE
        num_train_samples -= remaining

    train_indices = random.sample(indices, k=num_train_samples)
    test_indices = [index for index in indices if index not in train_indices]
    example_indices = random.sample(test_indices, k=NUM_EXAMPLES)
    x_train, y_train, x_test, y_test = process_graph_dataset(dataset, test_indices=test_indices)

    e.info('creating the model...')
    model: ks.models.Model = Model()
    model.compile(
        loss=[ks.losses.MeanSquaredError()],
        loss_weights=[1],
        metrics=[ks.metrics.MeanSquaredError()],
        optimizer=ks.optimizers.Adam(learning_rate=LEARNING_RATE),
        run_eagerly=False
    )

    with tf.device('gpu:0'):
        # NOTE: validation during the training is explicitly disabled because the validation dataset would
        # have to be divisible by the batch size as well. As that would be too much of a hassle we rather
        # disable it and perform the assessment manually afterwards.
        hist = model.fit(
            x_train,
            y_train,
            batch_size=BATCH_SIZE,
            epochs=EPOCHS,
            # validation_data=(x_test, y_test),
            # validation_freq=1,
            callbacks=LogProgressCallback(
                logger=e.logger,
                epoch_step=1,
                identifier=f'loss'
            ),
            verbose=0
        )
        e.info('finished model training')
        e.info(f'model with {model.count_params()} parameters')
        history = hist.history
        e['history'] = history

        e.info('calculating the gnnx explanations...')
        #out_test = y_test[0]
        #gnnx_importances(model, x=x_test, y=out_test, epochs=100)

        e.info('evaluating on the test set...')
        current = 0
        while current < len(test_indices):
            num_samples = min(len(test_indices) - current, EVAL_BATCH_SIZE)

            eval_indices = test_indices[current:current+num_samples]
            node_input = x_test[0][current:current+num_samples]
            edge_input = x_test[1][current:current+num_samples]
            edge_index_input = x_test[2][current:current+num_samples]

            out_true = y_test[0][current:current + num_samples]
            x_eval = (node_input, edge_input, edge_index_input)

            # ~ making predictions
            out_pred = model(x_eval)
            # ~ gnn explainer
            ni_pred, ei_pred = gnnx_importances(
                model,
                x_eval,
                out_pred,
                node_sparsity_factor=0.1,
                edge_sparsity_factor=0.1,
                learning_rate=0.01,
                epochs=250,
                logger=e.logger,
            )
            ni_pred = ni_pred.numpy()
            ei_pred = ei_pred.numpy()
            for c, index in enumerate(eval_indices):
                data = dataset_index_dict[index]
                g = data['metadata']['graph']
                e[f'out/pred/{index}'] = float(out_pred[c][0])
                e[f'out/true/{index}'] = float(out_true[c][0])
                e[f'ni/{index}'] = array_normalize(ni_pred[c])
                e[f'ei/{index}'] = array_normalize(ei_pred[c])

            current += num_samples
            e.info(f' * evaluated ({current}/{len(test_indices)})')

        e.info('calculating evaluation metrics...')
        errors = []
        ni_true = []
        ni_pred = []
        ei_true = []
        ei_pred = []
        for index in test_indices:
            data = dataset_index_dict[index]
            g = data['metadata']['graph']

            error = e[f'out/true/{index}'] - e[f'out/pred/{index}']
            errors.append(error)

            ni_pred += np.array(e[f'ni/{index}']).flatten().tolist()
            ni_true += np.array(g['node_importances']).flatten().tolist()

            ei_pred += np.array(e[f'ei/{index}']).flatten().tolist()
            ei_true += np.array(g['edge_importances']).flatten().tolist()

        mse = np.mean(np.square(errors))
        mae = np.mean(np.abs(errors))
        e['mse'] = mse
        e['mae'] = mae
        e.info(f' * mse: {mse}')

        r2 = r2_score(list(e['out/true'].values()), list(e['out/pred'].values()))
        e['r2'] = r2
        e.info(f' * r2: {r2}')

        ni_auc = roc_auc_score(ni_true, ni_pred)
        ei_auc = roc_auc_score(ei_true, ei_pred)
        e['ni_auc'] = ni_auc
        e['ei_auc'] = ei_auc
        e.info(f' * ni_auc: {ni_auc}')
        e.info(f' * ei_auc: {ei_auc}')

        e.info('drawing examples...')
        path = os.path.join(e.path, 'examples.pdf')
        num_channels = 1
        with PdfPages(path) as pdf:
            for c, index in enumerate(example_indices):
                ni = np.array(e[f'ni/{index}'])
                ei = np.array(e[f'ei/{index}'])

                data = dataset_index_dict[index]
                metadata = data['metadata']
                g = metadata['graph']
                g['node_coordinates'] = np.array(g['node_coordinates'])
                image = np.asarray(imread(data['image_path']))

                fig, rows = plt.subplots(ncols=2, nrows=num_channels, figsize=(16, 16),
                                         squeeze=False)

                for row_index in range(num_channels):
                    column_index = 0

                    # -- GROUND TRUTH --
                    ax_gt = rows[row_index][column_index]
                    ax_gt.imshow(image, extent=(0, image.shape[0], 0, image.shape[1]))
                    node_importances = np.array(g['node_importances'])
                    plot_node_importances(
                        g=g,
                        ax=ax_gt,
                        vmax=np.max(node_importances[:, row_index]),
                        node_importances=node_importances[:, row_index],
                        node_coordinates=g['node_coordinates']
                    )
                    edge_importances = np.array(g['edge_importances'])
                    plot_edge_importances(
                        g=g,
                        ax=ax_gt,
                        vmax=np.max(edge_importances[:, row_index]),
                        edge_importances=edge_importances[:, row_index],
                        node_coordinates=g['node_coordinates']
                    )
                    column_index += 1

                    # -- EXPLANATIONS --
                    ax_mas = rows[row_index][column_index]
                    ax_mas.imshow(image, extent=(0, image.shape[0], 0, image.shape[1]))
                    plot_node_importances(
                        g=g,
                        ax=ax_mas,
                        vmax=np.max(ni),
                        node_importances=ni[:, row_index],
                        node_coordinates=g['node_coordinates']
                    )
                    plot_edge_importances(
                        g=g,
                        ax=ax_mas,
                        vmax=np.max(ei),
                        edge_importances=ei[:, row_index],
                        node_coordinates=g['node_coordinates']
                    )
                    column_index += 1

                pdf.savefig(fig)
                plt.close(fig)

